static struct incoming_packet parse_packet(struct raw_packet packet)
{
  struct incoming_packet retval;
  unsigned char __far *ip_header = &packet.data[14];
  unsigned header_length;
  unsigned total_length;
  unsigned payload_len;
  unsigned char __far *payload_protocol;
  void __far *ptr;

  if(packet.length < 14)
  {
    retval.error = -4;
    return retval; // packet's length is less than normal ethernet header
  }

  #ifndef NOPROMISCUOUS
  if(_fmemcmp(packet.data, (void __far*)my_mac, 6) && _fmemcmp(packet.data, (void __far*)broadcast_mac, 6))
  {
    retval.error = -1;
    return retval; // this computer is not the destination
  }
  #endif

  if(packet.data[12] != 0x08)
  {
    retval.error = -2;
    return retval; // not an IP packet or an ARP packet
  }

  memset(&retval, 0, sizeof(struct incoming_packet));  

  #ifdef ARP
  if(packet.data[13] == 0x06) // this is an ARP packet
  {
    parse_arp(packet);
    return retval;
  }
  #endif

  if(packet.data[13])
  {
    retval.error = -7;
    return retval; // not an ipv4 or an ARP packet
  }

  if((ip_header[0] >> 4) != 4)
  {
    retval.error = -3;
    return retval; // not an ipv4 packet
  }

  _fmemcpy(retval.sender_mac, &packet.data[6], 6);

  header_length = (ip_header[0] & 0x0F) << 2;

  if(packet.length - 14 < header_length)
  {
    retval.error = -5;
    return retval; // packet length is less than header says
  }

  retval.ip_header.ipv4_header.type_of_service = ip_header[1];
  total_length = switch_endianness_s(*(unsigned short __far*)&ip_header[2]);

  if(packet.length - 14 < total_length)
  {
    retval.error = -6;
    return retval; // packet length is less than total length in header
  }

  retval.ip_header.ipv4_header.identification = switch_endianness_s(*(unsigned short __far*)&ip_header[4]);
  retval.ip_header.ipv4_header.flags.dont_fragment = ip_header[6] >> 6;
  retval.ip_header.ipv4_header.flags.more_fragments = ip_header[6] >> 5;
  retval.ip_header.ipv4_header.fragment_offset = switch_endianness_s(*(unsigned short __far*)&ip_header[6] & 0x9FFF);
  retval.ip_header.ipv4_header.time_to_live = ip_header[8];
  retval.ip_header.ipv4_header.protocol = ip_header[9];
  retval.ip_header.ipv4_header.src_address.p1 = ip_header[12];
  retval.ip_header.ipv4_header.src_address.p2 = ip_header[13];
  retval.ip_header.ipv4_header.src_address.p3 = ip_header[14];
  retval.ip_header.ipv4_header.src_address.p4 = ip_header[15];
  retval.ip_header.ipv4_header.dst_address.p1 = ip_header[16];
  retval.ip_header.ipv4_header.dst_address.p2 = ip_header[17];
  retval.ip_header.ipv4_header.dst_address.p3 = ip_header[18];
  retval.ip_header.ipv4_header.dst_address.p4 = ip_header[19];

  if(packet.data != loopback.packet.data)
  {
    if(ipv4_header_checksum(ip_header, header_length))
    {
      retval.error = -7;
      return retval;
    }
    if(retval.ip_header.ipv4_header.src_address.p1 == 127)
    { // drop incoming packets that are "from localhost"
      retval.error = -9;
      return retval;
    }
    if(lease.status <= 0)
      if(_fmemcmp(&retval.ip_header.ipv4_header.dst_address, &ipv4_my_addr, sizeof(struct ipv4_addr))
         && _fmemcmp(&retval.ip_header.ipv4_header.dst_address, &ipv4_broadcast, sizeof(struct ipv4_addr)))
    {
      retval.error = -8;
      return retval;
    }
  }

  payload_protocol = &ip_header[header_length];
  payload_len = total_length - header_length;

  switch(retval.ip_header.ipv4_header.protocol)
  {
    case 1:
      parse_icmp(payload_protocol, &retval.payload_header.icmp_header, &retval.data.icmp_echo, payload_len);
      retval.error = 1;
      break;
    case 6:
      parse_tcp4(payload_protocol, &retval.payload_header.tcp4_header, &retval.data.tcp4_data, payload_len, retval.ip_header.ipv4_header.src_address, retval.sender_mac);
      break;
    case 17:
      parse_udp4(payload_protocol, &retval.payload_header.udp4_header, &retval.data.udp_data, payload_len, retval.ip_header.ipv4_header.src_address, retval.ip_header.ipv4_header.dst_address, retval.sender_mac);
      break;
  }
 
  return retval;
}

static void parse_incoming_packet(unsigned packet)
{
  struct incoming_packet incoming_packet = parse_packet(raw_packet[packet]);
  unsigned new_startindex = raw_packet_startindex;

  if(incoming_packet.error == 1)
  {
    // respond to ping...
    if(is_icmp_echo_request(incoming_packet.payload_header.icmp_header))
      if(raw_packet[packet].length <= OUT_PACKET_SIZE)
        respond_ping(&incoming_packet);
  }
  else if(incoming_packet.error < 0)
    rx_error_count++;

  new_startindex += raw_packet[packet].length;
  new_startindex &= RAW_PACKET_BUFFER_MAX;
  raw_packet_startindex = new_startindex;
  raw_packet[packet].data = 0;
}

static void parse_loopback(void)
{
  loopback.processing = 1;
  loopback.has_data = 0;
  parse_packet(loopback.packet);
}
